import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:strings/strings.dart';

extension StringExtension on String {
  String lowerTitleCase() {
    return "${this[0].toLowerCase()}${this.substring(1)}";
  }
}

String collectionGenerator(FieldElement collectionField) {
  final stringBuffer = StringBuffer();
  final displayName = collectionField.displayName;

  if (!collectionField.isLate) {
    throw '$displayName must be `late`';
  }

  final collectionType = collectionField.type;

  if (collectionType is! InterfaceType) {
    throw '${collectionType.getDisplayString(withNullability: false)} must be an interface type';
  }

  final collectionTypes = collectionType.typeArguments;

  if (collectionTypes.length != 2) {
    throw StateError('Expected two types on FireCollectionReference');
  }

  final docName =
      collectionTypes.first.getDisplayString(withNullability: false);
  final modelName =
      collectionTypes.last.getDisplayString(withNullability: false);
  final lowerCaseModel = modelName.lowerTitleCase();

  final classTypeElement = collectionType.element;

  if (classTypeElement is! ClassElement) {
    throw '${classTypeElement.displayName} must be a class type of FireCollectionReference';
  }

  final className = capitalize(displayName);
  final suffix = RegExp('Collection', caseSensitive: false);
  final prefix = className.split(suffix)[0];
  final querySnapshotName = prefix + 'QuerySnapshot';

  stringBuffer.write('''
class $className extends FireCollectionReference<$docName, $modelName> {
  final CollectionReference _reference;

  $className(this._reference);

  FirebaseFirestore get firestore => _reference.firestore;

  String get id => _reference.id;

  DocumentReference? get parent {
    throw UnimplementedError();
  }

  String get path => _reference.path;

  Future<$docName> add($modelName $lowerCaseModel) async {
    final docReference = await _reference.add($lowerCaseModel.toJson());

    return $docName(
      data: $modelName.fromJson(docReference.data()!),
      reference: docReference, 
    );
  }

  UserDocumentReference doc([String? path]) {
    final documentReference = _reference.doc(path);
    return UserDocumentReference(documentReference);
  }

  Stream<$querySnapshotName> snapshots({bool includeMetadataChanges = false}) {
    return _reference
        .snapshots(includeMetadataChanges: includeMetadataChanges)
        .map((snapshot) => $querySnapshotName(snapshot));
  }

  Future<$querySnapshotName> get([GetOptions? getOptions]) async {
    final snapshot = await _reference.get(getOptions);

    return $querySnapshotName(snapshot);
  }
}
        ''');

  final queryName = prefix + 'Query';

  stringBuffer.write('''
class $querySnapshotName with FireQuerySnapshot<$docName, $modelName> {
  final QuerySnapshot _snapshot;
  
  $querySnapshotName(this._snapshot);

  List<$queryName> get docs {
    throw UnimplementedError();
  }

  List<FireDocumentChange<Model, Doc>> get docChanges {
    throw UnimplementedError();
  }

  SnapshotMetadata get metadata {
    return _snapshot.metadata;
  }
}
''');

  stringBuffer.write('''
class $queryName with FireQuery<$docName, $modelName> {
  final Query _query;
  
  $querySnapshotName(this._query);
}
''');

  return stringBuffer.toString();
}
