import 'dart:async';

import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:fireproof_annotation/fireproof_annotation.dart' show Fireproof;
import 'package:source_gen/source_gen.dart';

class FireproofGenerator extends GeneratorForAnnotation<Fireproof> {
  @override
  FutureOr<String> generateForAnnotatedElement(
    Element element,
    ConstantReader annotation,
    BuildStep buildStep,
  ) async {
    final buffer = StringBuffer();
    final classElement = element as ClassElement;
    final className = classElement.displayName;

    final prefix = '_\$';
    final abstractClassName = prefix + className;
    final factoryFunctionName = abstractClassName + 'FromSnapshot';

    final dataField = classElement.getField('data');

    if (dataField == null) {
      throw StateError('Class $className must have a field with a data field');
    }

    final dataFieldType = dataField.type;
    final dataFieldTypeName =
        dataFieldType.getDisplayString(withNullability: false);

    // Generates FromSnapshot function
    buffer.write('''
$className $factoryFunctionName(DocumentSnapshot snapshot) {
  final data = snapshot.data();

  if (!snapshot.exists) {
    throw StateError('Snapshot \$snapshot does not exist');
  }

  if (data == null) {
    throw StateError('No data for snapshot \$snapshot');
  }

  return $className(
    data: $dataFieldTypeName.fromJson(data),
    snapshot: snapshot,
  );
}

''');

    final docClassName = className + 'Doc';

    // Generates abstract FireDocumentSnapshot
    buffer.write('''
abstract class $abstractClassName implements FireDocumentSnapshot<$dataFieldTypeName, $docClassName> {
  final DocumentSnapshot _snapshot;

  final $docClassName reference;

  $abstractClassName(this._snapshot) : reference = $docClassName(_snapshot.reference);

  @override
  $dataFieldTypeName get data;

  @override
  String get id => reference.id;

  @override
  SnapshotMetadata get metadata => _snapshot.metadata;

  @override
  bool get exists => _snapshot.exists;

  @override
  dynamic get(dynamic field) {
    throw UnimplementedError();
  }
}

''');

    // Generates the concrete document reference
    buffer.write('''
class $docClassName with FireDocumentReference<$dataFieldTypeName, $docClassName> {
  final DocumentReference _reference;

  $docClassName(this._reference);

  FirebaseFirestore get firestore => _reference.firestore;

  @override
  String get id => _reference.id;

  @override
  // FireCollectionReference get parent => _reference.parent;
  FireCollectionReference get parent => throw UnimplementedError();

  @override
  String get path => _reference.path;

  @override
  FireCollectionReference<$dataFieldTypeName, $docClassName> collection(String collectionPath) {
    throw UnimplementedError();
  }

  @override
  Future<void> delete() async {
    await _reference.delete();
  }

  @override
  Future<$docClassName> get([GetOptions? options]) async {
    final documentSnapshot = await _reference.get(options);

    return $docClassName(documentSnapshot.reference);
  }

  @override
  Future<void> set($dataFieldTypeName data, [SetOptions? options]) async {
    await _reference.set(data.toJson(), options);
  }

  @override
  Stream<$docClassName> snapshots({bool includeMetadataChanges = false}) async* {
    yield* _reference
        .snapshots(includeMetadataChanges: includeMetadataChanges)
        .map(
      (documentSnapshot) {
        if (!documentSnapshot.exists) {
          throw StateError('DocumentSnapshot \$documentSnapshot does not exist');
        }

        return $docClassName(documentSnapshot.reference);
      },
    );
  }

  @override
  Future<void> update($dataFieldTypeName data) async {
    await _reference.update(data.toJson());
  }
}
''');

    return buffer.toString();
  }
}
